/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2012 Smithsonian Astrophysical Observatory
 *
 * This file is part of blur
 *
 * blur is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <math.h>

#include <randlib_c/randlib.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <bpipe/bpipe.h>
#include <bpipe/datatypes.h>
#include <bpipe/bpipe_axaf.h>

#include "main.h"

/* Routine to get the data pointer for a named data packet field */

static char *
dpktf_data( BPipe *bpipe, char *name, void *data )
{
  DpktField *field = bpipe_dpktf( bpipe, name );

  if ( field == NULL )
    tf_exit( ExitERR_misc, "%s isn't in input data stream", name);

  return bpipe_dpktf_data( field, data );

}

void
process(void)
{

  BPipe *bpipe;
  BPipeOutput *bpo;
  void  *data;
  size_t dpkt_core_size;
  size_t nphot;
  size_t n;
  DVector3 *position;

  /* for Gaussian multivariate routine */
  double parm[2*(2+3)/2+1];

  tf_enter("blur");

/* Setup bpipe input and output streams */

  if ( ( bpipe = bpipe_new() ) == NULL)
    tf_exit(ExitERR_misc, "couldn't open bpipe: %s",
	    bpipe_strerror(bpipe_errno));

  if ( ( bpipe_input(bpipe, c.input) ) )
    tf_exit(ExitERR_misc, "couldn't open input file `%s': %s", c.input,
	    bpipe_strerror(bpipe_errno));

  if ( NULL == ( bpo = bpipe_output( bpipe, c.output ) ) )
    tf_exit(ExitERR_misc, "couldn't open input file `%s': %s", c.output,
	    bpipe_strerror(bpipe_errno));

  if (NULL == (data = bpipe_map_alloc(bpipe,1, &dpkt_core_size)))
    tf_exit( ExitERR_alloc, "unable to allocate memory for data packet: %s",
	    bpipe_strerror(bpipe_errno));

  position = (DVector3 *) dpktf_data( bpipe, BP_AXAF_phot_pos, data);

  if ( bpipe_write_hdr(bpipe) )
    tf_exit(ExitERR_misc,"bpipe_write_hdr: %s",bpipe_strerror(bpipe_errno));

  /* initialize random number generator and get to correct block */

  setall(c.seed1, c.seed2);
  for ( n = 0 ; n < c.block ; n++ )
    initgn(1);

  /* set up gaussian engine for 2D gaussian */
  if ( FUNC_gauss1D == c.blurfunc )
  {
    double meanv[2] = { 0, 0 };
    double covm[4] = { 0, 0, 0, 0 };
    covm[0] = covm[3] = c.std_dev * c.std_dev;
    setgmn( meanv, covm, 2, parm );
  }

  /* Read-in bpipe photon data packets */

  bpipe_errno = BPNOERROR;
  nphot = 0;                    /* to keep track of each ray */

  while( bpipe_read_dpkts( bpipe, data, 1 )   == 1)
  {
    double q[2], work[2];

    switch(c.blurfunc)
    {
    case FUNC_gauss1D:
      genmn( parm, q, work );
      position->x += c.scale * q[0];
      position->y += c.scale * q[1];
      break;

    case FUNC_rect_uni:
      position->x += c.scale * genunf( -c.half_width,  c.half_width  );
      position->y += c.scale * genunf( -c.half_height, c.half_height );
      break;
    }

    nphot++;

    /* Write the data packet out  */

    if ( bpipe_write_dpkt( bpipe, data, bpo ) )
      tf_exit( ExitERR_misc, "error writing photon in blur %d", nphot + 1 );

  }   /* End of main loop by ray data packets */

  if ( bpipe_errno != BPNOERROR )
    tf_exit( ExitERR_misc, bpipe_strerror( bpipe_errno ) );

  bpipe_delete(bpipe);
  tf_leave();
}
